@@ -49,12 +49,9 @@
 
 #define noVERBOSE
 
-typedef struct _RsvgState RsvgState;
-typedef struct _RsvgSaxHandler RsvgSaxHandler;
-
 #define SVG_BUFFER_SIZE (1024 * 8)
 
-struct _RsvgState {
+typedef struct {
   double affine[6];
 
   gint opacity; /* 0..255 */
@@ -78,9 +75,11 @@ struct _RsvgState {
   gboolean in_defs;
 
   GdkPixbuf *save_pixbuf;
-};
+} RsvgState;
 
-struct _RsvgSaxHandler {
+typedef struct RsvgSaxHandler RsvgSaxHandler;
+
+struct RsvgSaxHandler {
   void (*free) (RsvgSaxHandler *self);
   void (*start_element) (RsvgSaxHandler *self, const xmlChar *name, const xmlChar **atts);
   void (*end_element) (RsvgSaxHandler *self, const xmlChar *name);
@@ -152,9 +151,9 @@ rsvg_ctx_free_helper (gpointer key, gpointer value, gpointer user_data)
 
   /* key == entval->name, so it's implicitly freed below */
 
-  g_free ((xmlChar *)entval->name);
-  g_free ((xmlChar *)entval->ExternalID);
-  g_free ((xmlChar *)entval->SystemID);
+  g_free ((char *) entval->name);
+  g_free ((char *) entval->ExternalID);
+  g_free ((char *) entval->SystemID);
   xmlFree (entval->content);
   xmlFree (entval->orig);
   g_free (entval);
@@ -177,8 +176,8 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
   RsvgState *state;
   gboolean has_alpha = 1;
   gint new_width, new_height;
-  gdouble x_zoom;
-  gdouble y_zoom;
+  double x_zoom;
+  double y_zoom;
 
   if (atts != NULL)
     {
@@ -189,16 +188,31 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 	  else if (!strcmp ((char *)atts[i], "height"))
 	    height = rsvg_css_parse_length ((char *)atts[i + 1], &fixed);
 	}
+
 #ifdef VERBOSE
       fprintf (stdout, "rsvg_start_svg: width = %d, height = %d\n",
 	       width, height);
 #endif
 
+      if (width == 0 || height == 0)
+        {
+          /* FIXME: GError here? */
+          g_warning ("rsvg_start_svg: can't render 0-sized SVG");
+          return;
+        }
+
       new_width = width;
       new_height = height;
       if (ctx->size_func)
 	(* ctx->size_func) (&new_width, &new_height, ctx->user_data);
 
+      if (new_width == 0 || new_height == 0)
+	{
+          /* FIXME: GError here? */
+          g_warning ("rsvg_start_svg: can't render 0-sized SVG");
+          return;
+	}
+
       x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
       y_zoom = (height < 0 || new_height < 0) ? 1 : (double) new_height / height;
 
@@ -208,20 +222,22 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 
       if (new_width < 0 || new_height < 0)
         {
-          g_warning ("rsvg_start_svg: no width and height attributes in SVG, nor supplied by size_func");
+          g_warning ("rsvg_start_svg: width and height not specified in the SVG, nor supplied by the size callback");
           if (new_width < 0) new_width = 500;
           if (new_height < 0) new_height = 500;
         }
 
       if (new_width >= INT_MAX / 4)
         {
-          /* FIXME: What warning, GError here? */
+          /* FIXME: GError here? */
+          g_warning ("rsvg_start_svg: width too large");
 	  return;
         }
       rowstride = (new_width * (has_alpha ? 4 : 3) + 3) & ~3;
       if (rowstride > INT_MAX / new_height)
         {
-          /* FIXME: What warning, GError here? */
+          /* FIXME: GError here? */
+          g_warning ("rsvg_start_svg: width too large");
 	  return;
         }
 
@@ -230,7 +246,8 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
       pixels = g_try_malloc (rowstride * new_height);
       if (pixels == NULL)
         {
-          /* FIXME: What warning, GError here? */
+          /* FIXME: GError here? */
+          g_warning ("rsvg_start_svg: dimensions too large");
 	  return;
         }
       memset (pixels, has_alpha ? 0 : 255, rowstride * new_height);
@@ -1613,45 +1630,83 @@ rsvg_handle_free (RsvgHandle *handle)
   g_free (handle);
 }
 
+typedef enum {
+  RSVG_SIZE_ZOOM,
+  RSVG_SIZE_WH,
+  RSVG_SIZE_WH_MAX,
+  RSVG_SIZE_ZOOM_MAX
+} RsvgSizeType;
+
 struct RsvgSizeCallbackData
 {
-  gdouble x_zoom;
-  gdouble y_zoom;
+  double x_zoom;
+  double y_zoom;
   gint width;
   gint height;
+  RsvgSizeType type;
   gboolean zoom_set;
   gboolean max_size_set;
 };
 
 static void
-rsvg_size_callback (gint     *width,
-		    gint     *height,
+rsvg_size_callback (int *width,
+		    int *height,
 		    gpointer  data)
 {
-  struct RsvgSizeCallbackData *real_data = (struct RsvgSizeCallbackData *)data;
-  gdouble zoomx, zoomy, zoom;
+  struct RsvgSizeCallbackData *real_data = (struct RsvgSizeCallbackData *) data;
+  double zoomx, zoomy, zoom;
 
-  if (real_data->zoom_set)
-    {
-      (* width) = floor (real_data->x_zoom * (* width) + 0.5);
-      (* height) = floor (real_data->y_zoom * (* height) + 0.5);
-    }
-  else if (real_data->max_size_set)
-    {
-      zoomx = (gdouble) real_data->width / (* width);
-      zoomy = (gdouble) real_data->height / (* height);
-      zoom = MIN (zoomx, zoomy);
-      
-      (* width) = floor (zoom * (* width) + 0.5);
-      (* height) = floor (zoom * (* height) + 0.5);
-    }
-  else
-    {
-      if (real_data->width != -1)
-	*width = real_data->width;
-      if (real_data->height != -1)
-	*height = real_data->height;
-    }
+  switch (real_data->type) {
+  case RSVG_SIZE_ZOOM:
+    if (*width < 0 || *height < 0)
+      return;
+
+    *width = floor (real_data->x_zoom * *width + 0.5);
+    *height = floor (real_data->y_zoom * *height + 0.5);
+    return;
+
+  case RSVG_SIZE_ZOOM_MAX:
+    if (*width < 0 || *height < 0)
+      return;
+
+    *width = floor (real_data->x_zoom * *width + 0.5);
+    *height = floor (real_data->y_zoom * *height + 0.5);
+    
+    if (*width > real_data->width || *height > real_data->height)
+      {
+	zoomx = (double) real_data->width / *width;
+	zoomy = (double) real_data->height / *height;
+	zoom = MIN (zoomx, zoomy);
+	
+	*width = floor (zoom * *width + 0.5);
+	*height = floor (zoom * *height + 0.5);
+      }
+    return;
+
+  case RSVG_SIZE_WH_MAX:
+    if (*width < 0 || *height < 0)
+      return;
+
+    zoomx = (double) real_data->width / *width;
+    zoomy = (double) real_data->height / *height;
+    zoom = MIN (zoomx, zoomy);
+    
+    *width = floor (zoom * *width + 0.5);
+    *height = floor (zoom * *height + 0.5);
+    return;
+
+  case RSVG_SIZE_WH:
+    if (*width < 0 || *height < 0)
+      return;
+
+    if (real_data->width != -1)
+      *width = real_data->width;
+    if (real_data->height != -1)
+      *height = real_data->height;
+    return;
+  }
+
+  g_assert_not_reached ();
 }
 
 /**
@@ -1710,8 +1765,7 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
     }
 
   handle = rsvg_handle_new ();
-  data.zoom_set = TRUE;
-  data.max_size_set = FALSE;
+  data.type = RSVG_SIZE_ZOOM;
   data.x_zoom = x_zoom;
   data.y_zoom = y_zoom;
 
@@ -1727,6 +1781,67 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
   return retval;
 }
 
+/**
+ * rsvg_pixbuf_from_file_at_zoom_with_max:
+ * @file_name: A file name
+ * @x_zoom: The horizontal zoom factor
+ * @y_zoom: The vertical zoom factor
+ * @max_width: The requested max width
+ * @max_height: The requested max heigh
+ * @error: return location for errors
+ * 
+ * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is scaled
+ * from the size indicated by the file by a factor of @x_zoom and @y_zoom. If the
+ * resulting pixbuf would be larger than max_width/max_heigh it is uniformly scaled
+ * down to fit in that rectangle.The caller must assume the reference to the
+ * returned pixbuf. If an error occurred, @error is set and %NULL is returned.
+ * 
+ * Return value: A newly allocated #GdkPixbuf, or %NULL
+ **/
+GdkPixbuf  *
+rsvg_pixbuf_from_file_at_zoom_with_max (const gchar  *file_name,
+					double        x_zoom,
+					double        y_zoom,
+					gint          max_width,
+					gint          max_height,
+					GError      **error)
+{
+  FILE *f;
+  char chars[SVG_BUFFER_SIZE];
+  gint result;
+  GdkPixbuf *retval;
+  RsvgHandle *handle;
+  struct RsvgSizeCallbackData data;
+
+  g_return_val_if_fail (file_name != NULL, NULL);
+  g_return_val_if_fail (x_zoom > 0.0 && y_zoom > 0.0, NULL);
+
+  f = fopen (file_name, "r");
+  if (!f)
+    {
+      /* FIXME: Set up error. */
+      return NULL;
+    }
+
+  handle = rsvg_handle_new ();
+  data.type = RSVG_SIZE_ZOOM_MAX;
+  data.x_zoom = x_zoom;
+  data.y_zoom = y_zoom;
+  data.width = max_width;
+  data.height = max_height;
+
+  rsvg_handle_set_size_callback (handle, rsvg_size_callback, &data, NULL);
+  while ((result = fread (chars, 1, SVG_BUFFER_SIZE, f)) > 0)
+    rsvg_handle_write (handle, chars, result, error);
+  rsvg_handle_close (handle, error);
+  
+  retval = rsvg_handle_get_pixbuf (handle);
+
+  fclose (f);
+  rsvg_handle_free (handle);
+  return retval;
+}
+
 /**
  * rsvg_pixbuf_from_file_at_size:
  * @file_name: A file name
@@ -1762,8 +1877,7 @@ rsvg_pixbuf_from_file_at_size (const gchar *file_name,
       return NULL;
     }
   handle = rsvg_handle_new ();
-  data.zoom_set = FALSE;
-  data.max_size_set = FALSE;
+  data.type = RSVG_SIZE_WH;
   data.width = width;
   data.height = height;
 
@@ -1814,8 +1928,7 @@ rsvg_pixbuf_from_file_at_max_size (const gchar     *file_name,
       return NULL;
     }
   handle = rsvg_handle_new ();
-  data.zoom_set = FALSE;
-  data.max_size_set = TRUE;
+  data.type = RSVG_SIZE_WH_MAX;
   data.width = max_width;
   data.height = max_height;
 
@@ -1832,3 +1945,4 @@ rsvg_pixbuf_from_file_at_max_size (const gchar     *file_name,
   return retval;
 }
 
+
